// src/Session.cc
// (C) 2005 - 2017 Philip Endecott
// This file is part of Anyterm; see http://anyterm.org/

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


#include "Session.hh"

#include <sstream>
#include <iomanip>

#include <unistd.h>
#include <signal.h>

#include "config.hh"
#include "html.hh"
#include "editscript.hh"
#include "Error.hh"
#include "Iconver.hh"
#include "unicode.hh"
#include "expand_command.hh"



Session::Session(int r, int c, int sb,
                 std::string host, std::string user, std::string param,
                 int t,
                 std::string command,
                 std::string charset_,
                 bool diff_):
  scrollback(sb),
  time_out(t),
  charset(charset_),
  diff(diff_),
  utf8_to_charset("UTF-8",charset),
  charset_to_ucs4(charset,UCS4_NATIVE),
  ucs4_to_utf8(UCS4_NATIVE,"UTF-8"),
  screen(r,c,scrollback),
  error(false),
  term(screen),
  subprocess( [this](std::string s) {this->process_output(s);},
              [this](std::string s) {this->process_error(s);},
              expand_command(command,host,user,param),
              r, c )
{
  dirty=true;
  touch();
}


Session::~Session()
{}


// Timeout unused sessions:

void Session::touch(void)
{
  last_access = time(NULL);
}


// Check if session has a backend error to report

void Session::report_any_backend_error(void)
{
  screen_lock_t l(screen_mut);
  if (error) {
    error=false;
    throw Error(error_msg);
  }
}


// Handle Apache requests:

std::string int_to_string(int i)
{
  char b[32];
  snprintf(b,sizeof(b),"%d",i);
  return b;
}


void Session::send(std::string k)
{
  if (!k.empty()) {
    std::string dk = utf8_to_charset(k);
    subprocess.send(dk);
  }

  touch();
}



std::string escape_html(std::string s)
{
  std::string t;
  for(std::string::size_type i=0; i<s.length(); i++) {
    char c=s[i];
    switch(c) {
    case '<': t+="&lt;"; break;
    case '>': t+="&gt;"; break;
    case '&': t+="&amp;"; break;
    default:  t+=c; break;
    }
  }
  return t;
}


std::string Session::rcv(void)
{
  {
    screen_lock_t l(screen_mut);
    if (!dirty && !error) {
      dirty_condition.wait_for(l, std::chrono::seconds(10));
    }
  }
  touch();

  report_any_backend_error();

  bool was_dirty;
  ucs4_string html_screen;
  {
    screen_lock_t l(screen_mut);
    was_dirty = dirty;
    dirty=false;
    if (!was_dirty) {
      return "n";
    }
    html_screen = htmlify_screen(screen);
  }

  std::string utf8_editscript;
  if (diff) {
    ucs4_string editscript = make_editscript(old_html_screen,html_screen);
    old_html_screen=html_screen;
    utf8_editscript = ucs4_to_utf8(editscript);
  } else {
    utf8_editscript = "R";
    utf8_editscript.append(ucs4_to_utf8(html_screen));
  }

  return utf8_editscript;
}


void Session::process_output(std::string s)
{
  ucs4_string ucs4_s = charset_to_ucs4(s);
  {
    screen_lock_t l(screen_mut);
    term.write(ucs4_s);
    dirty = true;
  }
  dirty_condition.notify_all();
}


void Session::process_error(std::string s)
{
  {
    screen_lock_t l(screen_mut);
    error_msg = s;
    error = true;
  }
  dirty_condition.notify_all();
}


bool Session::timed_out(void)
{
  return time(NULL) - last_access > time_out;
}


void Session::resize(int rows, int cols)
{
  {
    screen_lock_t l(screen_mut);
    screen.resize(rows,cols);
    dirty = true;
  }
  dirty_condition.notify_all();
  subprocess.resize(rows,cols);
}

